#define USE_KWIN

#include "kdialogd.h"
#include <iostream>
#include <kdiroperator.h>
#include <kapplication.h>
#include <qsocketnotifier.h>
#include <kio/netaccess.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <kconfig.h>
#include <kurlcombobox.h>
#ifdef USE_KWIN
#include <kwin.h>
#else
#include <X11/Xlib.h>
#endif
#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <ksockaddr.h>
#include <kdebug.h>
#include <kdeversion.h>

KConfig *cfg=NULL;
#define CFG_KEY_DIALOG_SIZE "KDialogDSize"

static QString groupName(const QString &app, bool fileDialog=true)
{
   return QString(fileDialog ? "KFileDialog " : "KDirSelectDialog ")+app;
}

// from kdebase/kdesu
static int createSocket()
{
    int         sockfd;
    ksocklen_t  addrlen;
    struct stat s;
    const char  *sock=getSockName(1);  // CPD TODO
    int         stat_err=lstat(sock, &s);

    if(!stat_err && S_ISLNK(s.st_mode))
    {
        kdWarning() << "KDialogD - Someone is running a symlink attack on you" << endl;
        if(unlink(sock))
        {
            kdWarning() << "KDialogD - Could not delete symlink" << endl;
            return -1;
        }
    }

    if (!access(sock, R_OK|W_OK))
    {
        kdWarning() << "KDialogD - stale socket exists" << endl;
        if (unlink(sock))
        {
            kdWarning() << "KDialogD - Could not delete stale socket" << endl;
            return -1;
        }
    }

    sockfd = socket(PF_UNIX, SOCK_STREAM, 0);
    if (sockfd < 0)
    {
        kdError() << "KDialogD - socket(): " << errno << endl;
        return -1;
    }

    struct sockaddr_un addr;
    addr.sun_family = AF_UNIX;
    strncpy(addr.sun_path, sock, sizeof(addr.sun_path)-1);
    addr.sun_path[sizeof(addr.sun_path)-1] = '\000';
    addrlen = SUN_LEN(&addr);
    if (bind(sockfd, (struct sockaddr *)&addr, addrlen) < 0)
    {
        kdError() << "KDialogD - KDialogD - bind(): " << errno << endl;
        return -1;
    }

    struct linger lin;
    lin.l_onoff = lin.l_linger = 0;
    if (setsockopt(sockfd, SOL_SOCKET, SO_LINGER, (char *) &lin, sizeof(linger)) < 0)
    {
        kdError() << "KDialogD - setsockopt(SO_LINGER): " << errno << endl;
        return -1;
    }
    
    int opt = 1;
    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *) &opt, sizeof(opt)) < 0)
    {
        kdError() << "KDialogD - setsockopt(SO_REUSEADDR): " << errno << endl;
        return -1;
    }
    opt = 1;
    if (setsockopt(sockfd, SOL_SOCKET, SO_KEEPALIVE, (char *) &opt, sizeof(opt)) < 0)
    {
        kdError() << "KDialogD - setsockopt(SO_KEEPALIVE): " << errno << endl;
        return -1;
    }

    if(::listen(sockfd, 1)<0)
    {
        kdError() << "KDialogD - listen(1): " << errno << endl;
        return -1;
    }

    //chmod(sock, 0600);
    return sockfd;
}

static void urls2Local(KURL::List &urls, QStringList &items, QWidget *parent)
{
    KURL::List::Iterator it(urls.begin()),
                         end(urls.end());

    for(; it!=end; ++it)
        if((*it).isLocalFile())
            items.append((*it).path());
        else
        {
#if KDE_IS_VERSION(3, 5, 0)
            KURL url(KIO::NetAccess::mostLocalURL(*it, parent));

            if(url.isLocalFile())
                items.append(url.path());
            else
                break;
#else
            break;
#endif
        }
}

KDialogD::KDialogD(QObject *parent)
        : QObject(parent),
          fd(::createSocket())
{
    if(fd<0)
        kdError() << "KDialogD - KDialogD could not create socket" << endl;
    else
    {
        if(!cfg)
            cfg=new KConfig("kdialogdrc", false, false);

        connect(new QSocketNotifier(fd, QSocketNotifier::Read, this),
                SIGNAL(activated(int)), this, SLOT(newConnection()));
    }
}

KDialogD::~KDialogD()
{
    if(-1!=fd)
        close(fd);
    if(cfg)
        delete cfg;
    cfg=NULL;
}

void KDialogD::newConnection()
{
    kdDebug() << "KDialogD - New connection" << endl;

    ksocklen_t         addrlen = 64;
    struct sockaddr_un clientname;
    int                connectedFD;

    if((connectedFD=::accept(fd, (struct sockaddr *) &clientname, &addrlen))>=0)
    {
        int appNameLen;

        if(readBlock(connectedFD, (char *)&appNameLen, 4))
        {
            bool     ok=true;
            QCString appName;

            if(0==appNameLen)
                appName="Generic";
            else
            {
                appName.resize(appNameLen);
                ok=readBlock(connectedFD, appName.data(), appNameLen);
            }

            if(ok)
                connect(new KDialogDClient(connectedFD, appName, this),
                        SIGNAL(error(KDialogDClient *)),
                        this, SLOT(deleteConnection(KDialogDClient *)));
        }
    }
}

void KDialogD::deleteConnection(KDialogDClient *client)
{
    kdDebug() << "KDialogD - Delete client" << endl;
    delete client;
}

KDialogDClient::KDialogDClient(int sock, const QString &an, QObject *parent)
              : QObject(parent),
                fd(sock),
                dlg(NULL),
                accepted(false),
                appName(an)
{
    kdDebug() << "KDialogD - new client..." << appName << " (" << fd << ")" << endl;
    connect(new QSocketNotifier(fd, QSocketNotifier::Read, this),
            SIGNAL(activated(int)), this, SLOT(read()));
}

KDialogDClient::~KDialogDClient()
{
    kdDebug() << "KDialogD - delete client..." << fd << endl;
    ::close(fd);
    if(cfg)
        cfg->sync();
}

void KDialogDClient::close()
{
    emit error(this);
    if(dlg)
    {
        dlg->close();
        dlg->delayedDestruct();
        dlg=NULL;
    }
}

void KDialogDClient::read()
{
    char         request;
    QString      caption;
    unsigned int xid=0;

    if(!dlg && readData(&request, 1) && request>=(char)OP_FILE_OPEN && request<=(char)OP_FOLDER &&
       readData((char *)&xid, 4) && readString(caption))
    {
        if(OP_FOLDER==(Operation)request)
        {
            QString intialFolder;

            if(readString(intialFolder))
            {
                initDialog(caption, new KDialogDDirSelectDialog(appName, intialFolder, true, NULL,
                           "folderdialog", false), xid);
                return;
            }
        }
        else
        {
            QString intialFolder,
                    filter;
            char    overW=0;

            if(readString(intialFolder) && readString(filter) &&
               (OP_FILE_SAVE!=(Operation)request || readData(&overW, 1)))
            {
                initDialog(caption, new KDialogDFileDialog(appName, (Operation)request, intialFolder,
                                                           filter, overW ? true : false), xid);
                return;
            }
        }
    }

    kdDebug() << "KDialogD - Comms error, closing connection..." << fd << endl;
    // If we get here something was wrong, close connection...
    close();
}

void KDialogDClient::finished()
{
    //
    // * finished is emitted when a dialog is ok'ed/cancel'ed/closed
    // * if the user just closes the dialog - neither ok nor cancel are emitted
    // * the dir select dialog doesnt seem to set the QDialog result parameter
    //   when it is accepted - so for this reason if ok is clicked we store an
    //   'accepted' value there, and check for that after the dialog is finished.
    kdDebug() << "KDialogD - finished" << endl;
    if(dlg && !(accepted || QDialog::Accepted==dlg->result()))
        cancel();
}

void KDialogDClient::ok(const QStringList &items)
{
    kdDebug() << "KDialogD - ok" << endl;

    int                        num=items.count();
    QStringList::ConstIterator it(items.begin()),
                               end(items.end());
    bool                       error=!writeData((char *)&num, 4);

    for(; !error && it!=end; ++it)
        error=!writeString(*it);

    if(error)
        close();
    else
        accepted=true;
    if(dlg)
        dlg->delayedDestruct();
    dlg=NULL;
}

void KDialogDClient::cancel()
{
    if(dlg)
    {
        kdDebug() << "KDialogD - cancel" << endl;

        int rv=0;

        if(!writeData((char *)&rv, 4))
            close();
        if(dlg)
            dlg->delayedDestruct();
        dlg=NULL;
    }
}

bool KDialogDClient::readData(QCString &buffer, int size)
{
    buffer.resize(size);
    return ::readBlock(fd, buffer.data(), size);
}

bool KDialogDClient::readString(QString &str)
{
    int size;

    if(!readData((char *)&size, 4))
        return false;

    QCString buffer;
    buffer.resize(size);

    if(!readData(buffer.data(), size))
        return false;

    str=QString::fromUtf8(buffer.data());
    return true;
}

bool KDialogDClient::writeString(const QString &str)
{
    QCString utf8(str.utf8());

    int size=utf8.length()+1;

    return writeData((char *)&size, 4) && writeData(utf8.data(), size);
}

void KDialogDClient::initDialog(const QString &caption, KDialogBase *d, unsigned int xid)
{
    accepted=false;
    dlg=d;

    if(!caption.isEmpty())
        dlg->setPlainCaption(caption);

    if(xid)
    {
#ifdef USE_KWIN
        KWin::setMainWindow(dlg, xid);
        KWin::setState(dlg->winId(), NET::Modal);

        KWin::WindowInfo wi(KWin::windowInfo(xid, NET::WMGeometry, NET::WM2UserTime));
        QRect            geom(wi.geometry());
        int              rx=geom.x(),
                         ry=geom.y();

        rx=(rx+(geom.width()/2))-(dlg->width()/2);
        if(rx<0)
            rx=0;
        ry=(ry+(geom.height()/2))-(dlg->height()/2);
        if(ry<0)
            ry=0;
        dlg->move(rx, ry);
#else
        XWindowAttributes attr;
        int               rx, ry;
        Window            junkwin;

        XSetTransientForHint(qt_xdisplay(), dlg->winId(), xid);
        if(XGetWindowAttributes(qt_xdisplay(), xid, &attr))
        {
            XTranslateCoordinates(qt_xdisplay(), xid, attr.root,
                                  -attr.border_width, -16,
                                  &rx, &ry, &junkwin);

            rx=(rx+(attr.width/2))-(dlg->width()/2);
            if(rx<0)
                rx=0;
            ry=(ry+(attr.height/2))-(dlg->height()/2);
            if(ry<0)
                ry=0;
            dlg->move(rx, ry);
        }
#endif
    }

    connect(dlg, SIGNAL(ok(const QStringList &)), this, SLOT(ok(const QStringList &)));
    connect(dlg, SIGNAL(finished()), this, SLOT(finished()));
    dlg->show();
}

KDialogDFileDialog::KDialogDFileDialog(QString &an, Operation op, const QString &startDir,
                                       const QString &filter, bool confirmOw)
                  : KFileDialog(startDir.isEmpty() || "~"==startDir ? QDir::homeDirPath() : startDir,
                                filter, NULL, NULL, false),
                    itsConfirmOw(confirmOw),
                    appName(an)
{
    kdDebug() << "KDialogD - startDir:" << startDir << endl;

    switch(op)
    {
        case OP_FILE_OPEN:
            setOperationMode(KFileDialog::Opening);
            setMode((KFile::Mode)(KFile::File | KFile::ExistingOnly));
            break;
        case OP_FILE_OPEN_MULTIPLE:
            setOperationMode(KFileDialog::Opening);
            setMode((KFile::Mode)(KFile::Files | KFile::ExistingOnly));
            break;
        case OP_FILE_SAVE:
            setOperationMode(KFileDialog::Saving);
            setMode(KFile::File);
            break;
        default:
            break;
    }

    if(cfg)
    {
        QString oldGrp(cfg->group()),
                grp(groupName(appName));
        QSize   defaultSize(600, 400);

        readConfig(cfg, grp);
        cfg->setGroup(grp);
        resize(cfg->readSizeEntry(CFG_KEY_DIALOG_SIZE, &defaultSize));
        cfg->setGroup(oldGrp);
    }

    ops->clearHistory();
}

void KDialogDFileDialog::accept()
{
    kdDebug() << "KDialogD - KDialogDFileDialog::accept" << endl;
}

void KDialogDFileDialog::slotOk()
{
    setResult(QDialog::Accepted);
    KFileDialog::slotOk();

    kdDebug() << "KDialogD - KDialogDFileDialog::slotOk" << endl;
    KURL::List  urls;
    QStringList items;
    bool        good=true;

    if(mode()&KFile::Files)
        urls=selectedURLs();
    else if(!locationEdit->currentText().isEmpty())
        urls.append(selectedURL());

    if(urls.count())
    {
        urls2Local(urls, items, this);

        if(urls.count()!=items.count())
        {
            KMessageBox::error(this, i18n("You can only access local files."));
            good=false;
        }
        else if(itsConfirmOw && KFileDialog::Saving==operationMode())
            good=!KIO::NetAccess::exists(urls.first(), false, this) ||
                 KMessageBox::Continue==KMessageBox::warningContinueCancel(this,
                                            i18n("File %1 exits.\nDo you want to replace it?")
                                                 .arg(urls.first().prettyURL()),
                                            i18n("File Exists"),
                                            KGuiItem(i18n("Replace"), "filesaveas"), QString::null,
                                            KMessageBox::Notify|KMessageBox::PlainCaption);

        if(good)
        {
            QString filter(currentFilter());

            if(!filter.isEmpty())
                items.append(filter);

            emit ok(items);
            hide();
            KFileDialog::accept();
       }
       else
            setResult(QDialog::Rejected);
    }
}

KDialogDFileDialog::~KDialogDFileDialog()
{
    kdDebug() << "KDialogD - ~KDialogDFileDialog" << endl;

    if(cfg)
    {
        QString oldGrp(cfg->group()),
                grp(groupName(appName));

        writeConfig(cfg, grp);
        cfg->setGroup(grp);
        cfg->writeEntry(CFG_KEY_DIALOG_SIZE, size());
        cfg->setGroup(oldGrp);
    }
}

KDialogDDirSelectDialog::KDialogDDirSelectDialog(QString &an, const QString &startDir, bool localOnly,
                                                 QWidget *parent, const char *name, bool modal)
                       : KDirSelectDialog(startDir.isEmpty() || "~"==startDir
                                              ? QDir::homeDirPath() : startDir,
                                          localOnly, parent, name, modal),
                         appName(an)
{
    kdDebug() << "KDialogD - startDir:" << startDir << endl;

    if(cfg)
    {
        QString oldGrp(cfg->group()),
                grp(groupName(appName, false));
        QSize   defaultSize(600, 400);

        //readConfig(cfg, grp);
        cfg->setGroup(grp);
        resize(cfg->readSizeEntry(CFG_KEY_DIALOG_SIZE, &defaultSize));
        cfg->setGroup(oldGrp);
    }
}

KDialogDDirSelectDialog::~KDialogDDirSelectDialog()
{
    kdDebug() << "KDialogD - ~KDialogDDirSelectDialog" << endl;

    if(cfg)
    {
        QString oldGrp(cfg->group()),
                grp(groupName(appName, false));

        //writeConfig(cfg, grp);
        cfg->setGroup(grp);
        cfg->writeEntry(CFG_KEY_DIALOG_SIZE, size());
        cfg->setGroup(oldGrp);
    }
}

void KDialogDDirSelectDialog::slotOk()
{
    kdDebug() << "KDialogD - KDialogDDirSelectDialog::slotOk" << endl;

    KURL::List  urls;
    QStringList items;

    urls.append(url());
    urls2Local(urls, items, this);

    if(urls.count()!=items.count())
        KMessageBox::error(this, i18n("You can only access local folders."));
    else
    {
        emit ok(items);
        hide();
    }
}

extern "C"
{
    KDE_EXPORT KDEDModule *create_kdialogd(const QCString &obj)
    {
        return new KDialogDKDED(obj);
    }
};

KDialogDKDED::KDialogDKDED(const QCString &obj)
            : KDEDModule(obj)
{
    new KDialogD(this);
}

#include "kdialogd.moc"

